%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PNK: COMPACTING-GC.S
% Title:          Compacting garbage collector for PSL
% Author:         Martin Griss and Eric Benson
% Created:        28 August 1981
% Status:         Experimental
% Mode:           Lisp
%
% (c) Copyright 1987, 1988  Konrad Zuse-Zentrum Berlin, for modifications
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% 6-January-1988 W. Neun, ZIB Berlin
%   GC must preserve vector registers across its work, because some
%   silly fast-big routines rely on that!
% 25-Aug-1988  W. Neun , ZIB Berlin.
%   Redesign necessary: Gc is too slow on 'fat' machines even with partial
%   improvements described below. Speed up will use vector operations!
% 20-Aug-1988  W. Neun ,ZIB Berlin
%   This garbage collector is useable up to 256 M words now, which Cray-2
%   will use. Added some comments and some refinements in
%   inspect-oldstable, buildrelocationfields, updateheap and compactheap
%   in order to get gc faster.
% 23-Apr-1988  W. Neun, ZIB Berlin
%   Included and refined the compacti-extension technique.
%   Background are the new machines with larger memories at the
%   horizont and the need to address more than 16 M words of heap
%   Gc is now able to handle a minimum of 256 M words in one pass
% 23-Feb-1988 (W. Neun, ZIB Berlin)
%   Changed method for clearing the gcfields.
%   Implemented method for scanning stable part in order to allow
%   'Small' garbage collections.
% 6-Nov-1987 (W. Neun, ZIB Berlin)
%   Fixed code for heap sizes above 8 Mega words of heap Massive cleanup
%   of memory accesses, some lap coding;  Removed Clearstable pass
% 6-Dec-1986 (H. Melenk, ZIB Berlin)
%   Invented sweep-stack to cleanup the stack before gc is running
%   The actual contents of S and T Register may contain arbitrary
%   binary patterns, fooling the tag interpretation.
% 18-Aug-1986 (H.Melenk,ZIB Berlin)
%   inserted hooks for the handling of offsets larger than gcfield
%   capacity. COMPACTI-EXTENSION to be loaded on request if heap
%   exceeds 400000 words
% 7-Aug-1986 (W.Neun ,ZIB Berlin)
%   inserted calls to set-continue-at-attention to avoid exiting lisp
%   when garbage collector works and an attention interrupt occurs
% 29-May-1986 (W.Neun ,ZIB Berlin)
%   revised prereclaim and postreclaim to be automatically generated
% 15-May-1986 (H.Melenk ZIB Berlin)
%   clearmovementwithinsegment now done for stable part too
%   in function clearstable added functions prereclaim and postreclaim
%   to save local registers across garbage collection
% 24-April-1986 (H.Melenk, ZIB Berlin)
%  correction: Clearmovewithinsegment now inserts correct bitstring for
%   negint items
% 22-April-1986 (H.Melenk, ZIB Berlin)
%  modification of markfrombase in order to do cdr iteration
%  reduces stack requirement)
% 11-April-1986 (W. Neun, ZIB Berlin)
%  Interface routine for *CONS Cmacro (Cons-Reclaim)
% March-1986 (H. Melenk, ZIB Berlin)
%  Optimization for CRAY (register usage).
%  Reallocation routines for heap movement
% Feb-1986 (H. Melenk, ZIB Berlin)
%  Debugging for PSL 3.4 usage; implementation of an octal dump
%  in case of GC failure.
%  Inhibit system continue in case of GC failure (variable noreprieve*).
%
% 3-Jul-1984 (Vicki O'Day)
%  Changed references to the constant forward to forward-tag, to match
%  the declaration in tags
% 01-Dec-83 14:24:38 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
% WARNING!  This file has not been parameterized using
% AddressingUnitsPerItem.  It will not work on machines that
% address bytes /csp 3-1-83
% All data types have either explicit header tag in first item,
% or are assumed to be 1st element of pair.
% Revision History:
% Edit by Cris Perdue, 16 Feb 1983 1407-PST
% Fixed GtHeap and collector(s) to use only HeapLast, not HeapPreviousLast
% Sets HeapTrapped to NIL now.
% Using known-free-space function
%  Added check of Heap-Warn-Level after %Reclaim
%  Defined and used known-free-space function
%  <PSL.KERNEL>COMPACTING-GC.RED.9,  4-Oct-82 17:59:55, Edit by BENSON
%  Added GCTime!*
%  <PSL.KERNEL>COMPACTING-GC.RED.3, 21-Sep-82 10:43:21, Edit by BENSON
%  Flagged most functions internal
% (M.L. Griss, March, 1977).
% (Update to speed up, July 1978)
% Converted to Syslisp July 1980
% En-STRUCT-ed, Eric Benson April 1981
% Added EVECT tag, M. Griss, 3 July 1982
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Modifications of Compacting gc done in ZIB Berlin:
%
% Large memory sizes:
%
% This gc was designed when Cray machines had maximal memory sizes of
% 4 M words which can be easily handled by the gc field of 22 bits.
% Meanwhile X-MP machines have maximum memory of 16 M words, Cray Y-MP
% has memory size of 32 M words and Cray-2 has 256 M words.
% So the gc field of an item which has to hold the movelength is not
% sufficient anymore. To solve that problem in the buildreloactionfields
% process the overflow situation (in case of collapsing heap) is handled
% such that the location where overflow happend is stored into a vector
% and the gc field contains the overflowing part only.
% The vector information is inspected at update time. Since the vector
% is kept in a vector register this is sufficient for memory sizes up to
% 256 M words and can be easily enlarged by a second register.
%
% Another problem with X-MP machines with more than 8 M words of heap:
% The heappointers kept in B registers are 'negative' in sense of hardware
% and inline comparison in cons macro is wrong. So the heap is divided into
% two parts, the one below 8M and the upper one. Reclaim 'opens' the upper
% part when lower part is full so this garbage collection is a 'dummy', just
% filling the gap around 8M. Furthermore gc has to choose the appropriate
% pointer sets after reclaim :
% If heaplast is <8M then heaptrapbound is set to 8M and to
% heapupperbound-2 if heaplast is above 8M.
%
% Small garbage collections
% Open compiled cons
%     to be done
%
 
% Utilities for testing and error cases : octal print and dump
 
    % the following fluid variables are held in T-registers
    % during garbage collection:  localhole localholeupper
 
(compiletime (progn
  (fluid '(localhole localholeupper
           localheapshrinkupper localheapshrinklower))
  (put 'localhole      'Treg 8#67)
  (put 'localholeupper 'Treg 8#66)
  (put 'localhole      'registername "T67")
  (put 'localholeupper 'registername "T66")
 
  (flag '(p&o p&m n&l putheapdata) 'internalfunction)
  (load hash-decls if-system vector-instructions inum
        sys-consts fast-vector )))
 
(de p&o(X) (putoct X) (terminalwritechar 32 32))
(de p&m(X)
   (cond  ((wlessp X 0) (putoct -1) (terminalwritechar 32 32))
          (t  (putoct (getmem X)) (terminalwritechar 32 32))))
(de n&l() (console-newline))
(de center-dump(n l)    % central dump
  (when *dump           % only if dump indicator is on
   (progn  (p&o n)(console-print-string " : center dump")
      %    (putheapdata n)
           (ifor (from i (wdifference n l)(wplus2 n l) 3)
             (do  (progn  (p&o i) (p&m  i) (p&m (wplus2 i 1))
                          (p&m (wplus2 i 2)) (n&l)
           ) )    )
           (n&l)
) ) )
 
(fluid  '(Bregsave Tregsave &oldstable&))
(global '(*small-garbage-collection*))
 
(de putheapdata (n)         % printing some diagnostic data
 
 (if *small-garbage-collection* (console-print-string "small gc")
                                (console-print-string "full gc"))
 %(backtrace) (putoct n)    (console-print-string " Argument")
  (putoct n)    (console-print-string " Argument")
  (putoct heaplowerbound)   (console-print-string " heaplowerbound")
  (putoct heapupperbound)   (console-print-string " heapupperbound")
  (putoct heaplast)         (console-print-string " heaplast")
  (putoct heaptrapbound)    (console-print-string " heaptrapbound")
  (putoct localhole)        (console-print-string " localhole")
  (putoct localholeupper)   (console-print-string " localholeupper")
  (putoct localheapshrinkupper) (console-print-string " localheapshrinkupper")
  (putoct localheapshrinklower) (console-print-string " localheapshrinklower")
  (putoct schaedling)           (console-print-string " schaedling")
  (putoct schaedling1)          (console-print-string " schaedling1")
  (putoct schaedling2)          (console-print-string " schaedling2")
  (putoct schaedling1a)         (console-print-string " schaedling1a")
  (putoct schaedling2a)         (console-print-string " schaedling2a")
  (when (eq schaedling 'region)
        (setq *dump t)
        (center-dump schaedling1 100))
  (when (eq schaedling 'heap)
        (setq *dump t)
        (center-dump n 100)
        (center-dump schaedling1 100))
  (putoct (car &oldstable&)) (terminalwritechar 32 32)
  (putoct (cdr &oldstable&)) (console-print-string " &oldstable&"))
 
(fluid '(*gc            % Controls printing of statistics
         gctime*        % Total amount of time spent in GC
         gcknt*         % count of # of GC's since system build
         heap-warn-level% Enlarging heap if this much not free after %Reclaim.
         *gc-large-move % a move of heapspace larger than critical-
                        % offset occurred
         Tregsave       % for scratch
         gcextrasave )) % savearea for special values in registers
 
(setq gcextrasave (mkvect 10))
 
(loadtime
  (progn (setq *gc t)       % Do print GC messages (SL Rep says no)
         (setq gctime* 0)
         (setq gcknt* 0)    % Initialize to zero
         (setq heap-warn-level 1000)
         nil))
 
(fluid '(heaptrapped heaplast heapupperbound heaplowerbound
         *noreprieve *dump *errout))
 
(compiletime (progn
 
 (flag
  '(markfromallbases compactheap markfromonesymbol updateheapupper
    gcmessage searchvec markfromsymbols markfromrange markfrombase
    gcerror markfromvector prereclaim postreclaim updateregion
    updateheap updatevectorregisters inspect-oldstable compactheapupper
    !%!&reclaim cons-reclaim-aux markfrombaseaux buildrelocationfields1
    buildrelocationfieldsupper buildrelocationfieldsupper1
    last-string-in-symnam last-noninteger-in-symnam makeidfreelist
    updatesymbols updateallbases
    upper-inspect-oldstable buildrelocationfields)
   'internalfunction)
 
% the following code is a little bit more friendly since it uses
% 1 (instead of 2 ) jumps and does calculation instead
 
(ds pointertagp (x)      %  x is a tag, which should be in range 1 .. 14
      (wgeq (wor (wdifference 14 x) (wdifference x 1)) 0 ))
 
(ds pointertagpp(x)     %same but without wgeq 0 ..
      (wor (wdifference 14 x) (wdifference x 1)) )
 
(ds withinheappointer (x)   %  x is a pointer, which should point to heap
    (wgeq (wor (wdifference localhole x)
               (wdifference x heaplowerbound))
          0))
 
 (define-constant gcmarkvalue 511)
 (define-constant critical-offset (expt 2 16))
 (define-constant critical-offset-bitsize 16)
 (define-constant hskip (compiler-constant 'forward-tag))
 
 (ds mark (x)      % Get GC mark bits in item X points to
          (gcfield (getmem x)))
 
 (ds signedinf (u) (signedfield u infstartingbit infbitlength))
 
              % insert a mark field into a local quantity
 (deflist
       '((insmark ((*putfield (reg 2) (reg 1) (wconst gcstartingbit)
                                              (wconst gcbitlength))))
 
               % insert an inf-field into a local quantity
         (insinf  ((*putfield (reg 2) (reg 1) (wconst infstartingbit)
                                              (wconst infbitlength))))
        ) 'Opencode)
 
  (put 'insmark 'destroys '((reg 1) (reg 2)))
 
  (put 'insinf 'destroys '((reg 1) (reg 2)))
 
  (ds setmark (x)    % Set GC mark bits in item X points to
             (setf (gcfield (getmem x)) gcmarkvalue))
 
  (ds marked (x)   % Is item pointed to by X marked?
              (weq (mark x) gcmarkvalue))
 
  (ds markid (x)
          (setf (field (symnam x) tagstartingbit tagbitlength)
                  forward-tag))
 
  (ds markedid (x) (weq (tag (symnam x)) forward-tag))
 
% local variants of markid and markedid, supplied with symnam entry
% this is useful since we got or will use the symnam entry in almost any
% case some instruction before we call markid or markedid
 
  (ds markIDinSymnam (x sx)
          (putfield sx tagstartingbit tagbitlength forward-tag)
          (putmem (aplus2 symnam x) sx))
 
  (ds markedIDinSymnam (sx)
          (weq (tag sx) forward-tag))
 
  (ds clearidmark (x)
         (setf (field (symnam x) tagstartingbit tagbitlength)
                    string-tag))
 
         % Relocation primitives
 
  (ds skiplength (x) % Stored in heap header
             (inf (getmem x)))
 
  (ds putskiplength (x l) % Store in heap header
             (setf (inf (getmem x)) l))
 
  (put 'skiplength 'assign-op 'putskiplength)
 
%reloc with test
 
 (ds reloc (x)                   % relocate the inf field of item X
    (prog (&&mem&& &&xx&&)       % and error test (temporary)
 
       (setq &&mem&& (getmem x))
       (setq &&xx&& (wrshift x critical-offset-bitsize))
 
       (when (wneq 0 &&xx&&)   % possible critical update
             (setq &&xx&& (searchvec x)))
 
       (setq &&mem&& (gcfield &&mem&&))
       (when (not (eq (wrshift &&mem&& critical-offset-bitsize)
                      (wrshift &&xx&& critical-offset-bitsize)))
           (putoct &&xx&&) (putoct &&mem&&) (putoct (searchvec x))(putoct x)
                  (console-print-string " maessige relocation"))
 
       (when (wlessp &&mem&& 1023) % test for illegal Reloc's
              (putoct &&mem&&)     % if = 511, illegal reference
                                   % if = 0 ,  outside unstable heap
              (console-print-string " Heulen & Zaehneklappern")
              (center-dump x 64)
              (return (putheapdata x)))
 
       (return  &&mem&&)))
%      (return (wplus2 &&xx&& &&mem&&))))
 
  (ds untested-reloc (x)              % Reloc field in item
         (if (wneq 0 (wrshift x critical-offset-bitsize))
             (wplus2 (searchvec x) (gcfield (getmem x)))
             (gcfield (getmem x))))
 
  (ds within-nonstable-heap (ptr)   % ptr points into nonstable heap
                                  % breaking test down to one sign test
     (wgreaterp (wor (wdifference ptr localhole)
                     (wdifference localholeupper ptr)) 0))
 
  (ds noncritcal-reloc (x)  % Not used by now
             (gcfield (getmem x)))
 
  (ds putmovementwithinsegment (x m) % Store reloc field
             (setf (gcfield (getmem x)) m))
 
  (ds clearmovementwithinsegment (x) % Clear reloc field
             (setf (gcfield (getmem x))
                   (if (negintp (getmem x)) 8#17777777 0)))
 
  (put 'MovementWithinSegment 'AssignOp 'PutMovementWithinSegment)
 
 nil))
 
(fluid '(stacklowerbound st savereprieve stackstart stackend
             % next ID number to be allocated and bindingstackpointer
         nextsymbol bndstkptr bndstklowerbound startingrealtime))
 
(setq &oldstable& (cons 0 0)) % car is oldstable of lower heap
                              % cdr dito of upper part
 
(compiletime   % need to clear the frame some times
 (if_system Cray-2
  (put 'clear-frame 'opencode
     '((*Move 0 (reg s6))         (*move (reg s6) (frame 1))
       (*move (reg s6) (frame 2)) (*move (reg s6) (frame 3))
       (*move (reg s6) (frame 4)) (*move (reg s6) (frame 5))
       (*move (reg s6) (frame 6)) (*move (reg s6) (frame 7))
       (*move (reg s6) (frame 8)) (*move (reg s6) (frame 9))
       (*move (reg s6) (frame 10))(*move (reg s6) (frame 11))
       (*move (reg s6) (frame 12))(*move (reg s6) (frame 13))
       (*move (reg s6) (frame 14))(*move (reg s6) (frame 15))
       (*move (reg s6) (frame 16))(*move (reg s6) (frame 17))
       (*move (reg s6) (frame 18))(*move (reg s6) (frame 19))
       (*move (reg s6) (frame 20))(*move (reg s6) (frame 21))
       (*move (reg s6) (frame 22))(*move (reg s6) (frame 23))
       (*move (reg s6) (frame 24))(*move (reg s6) (frame 25))
       (*move (reg s6) (frame 26))(*move (reg s6) (frame 27))
       (*move (reg s6) (frame 28))(*move (reg s6) (frame 29))
       (*move (reg s6) (frame 30))(*move (reg s6) (frame 31))
       (*move (reg s6) (frame 32))(*move (reg s6) (frame 33))
       (*move (reg s6) (frame 34))(*move (reg s6) (frame 35))
       (*move (reg s6) (frame 36))(*move (reg s6) (frame 37))
       (*move (reg s6) (frame 38))(*move (reg s6) (frame 39))
       (*move (reg s6) (frame 40))(*move (reg s6) (frame 41))
       (*move (reg s6) (frame 42))(*move (reg s6) (frame 43))
       (*move (reg s6) (frame 44))(*move (reg s6) (frame 45))
       (*move (reg s6) (frame 46))(*move (reg s6) (frame 47))
       (*move (reg s6) (frame 48))(*move (reg s6) (frame 49))
       (*move (reg s6) (frame 50))(*move (reg s6) (frame 51))
       (*move (reg s6) (frame 52))(*move (reg s6) (frame 53))
       (*move (reg s6) (frame 54))(*move (reg s6) (frame 55))
       (*move (reg s6) (frame 56))(*move (reg s6) (frame 57))
       (*move (reg s6) (frame 58))(*move (reg s6) (frame 59))
       (*move (reg s6) (frame 60))(*move (reg s6) (frame 61))
       (*move (reg s6) (frame 62))(*move (reg s6) (Reg t125))
       (*move (reg s6) (Reg t126))(*move (reg s6) (Reg t127))
       (*move (reg s6) (Reg t130))(*move (reg s6) (Reg t131))
       (*move (reg s6) (Reg t132))(*move (reg s6) (Reg t133))
       (*move (reg s6) (Reg t134))(*move (reg s6) (Reg t135))
       (*move (reg s6) (Reg t136))(*move (reg s6) (Reg t137))
       (*move (reg s6) (Reg t140))))
   (put 'clear-frame 'opencode
     '((*Move 0 (reg s6))         (*move (reg s6) (frame 1))
       (*move (reg s6) (frame 2)) (*move (reg s6) (frame 3))
       (*move (reg s6) (frame 4)) (*move (reg s6) (frame 5))
       (*move (reg s6) (frame 6)) (*move (reg s6) (frame 7))
       (*move (reg s6) (frame 8)) (*move (reg s6) (frame 9))
       (*move (reg s6) (frame 10))(*move (reg s6) (frame 11))
       (*move (reg s6) (frame 12))(*move (reg s6) (frame 13))
       (*move (reg s6) (frame 14))(*move (reg s6) (frame 15))
       (*move (reg s6) (frame 16))(*move (reg s6) (frame 17))
       (*move (reg s6) (frame 18))(*move (reg s6) (frame 19))
       (*move (reg s6) (frame 20))(*move (reg s6) (frame 21))
       (*move (reg s6) (frame 22))(*move (reg s6) (frame 23))
       (*move (reg s6) (frame 24))(*move (reg s6) (frame 25))
       (*move (reg s6) (frame 26))(*move (reg s6) (frame 27))
       (*move (reg s6) (frame 28))(*move (reg s6) (frame 29))
       (*move (reg s6) (frame 30))(*move (reg s6) (frame 31))
       (*move (reg s6) (frame 32))(*move (reg s6) (frame 33))
       (*move (reg s6) (Reg t71)) (*move (reg s6) (Reg t72))
       (*move (reg s6) (Reg t73)) (*move (reg s6) (Reg t74))
       (*move (reg s6) (Reg t75)) (*move (reg s6) (Reg t76))))
))
 
(de reclaim ()                 % User call to garbage collector
  (progn (clear-extraregs)
         (clear-frame)
         (!%reclaim 0 0 0 0 0) % clear all Lisp registers
         nil))
 
% Introducing the method of small garbage collection:
% The idea is that most references from stable part are pointing into
% stable part too, so they are not changed in gc. We will try to set
% heaplowerbound such that the stable part is not treated as 'in heap'.
% Exceptions in the stable part (i.e. pointers into not stable part)
% are written into stack, updated there and rewritten afterwards.
 
(de !%reclaim ()          % frame for small gc
  (prog (oldheaplowerbound small-variant oldstack numberofexceptions adr
         oldheapupperbound iii nothing-upper nothing-lower)
        (setq *small-garbage-collection* NIL)
        (setq oldheapupperbound heapupperbound
              oldheaplowerbound heaplowerbound )
 
        (setq startingrealtime (timc))
 
        (prereclaim) % put register values into gcextrasave array
                     % CAUTION:  IO is not allowed after update phase
                     %           until postreclaim is done
 
        (when (and (not (eq 0 (wand gcknt* 31))) % each 32th is a full gc
                   (wgreaterp (wplus2 (car &oldstable&) % trivial cases!!
                                      (cdr &oldstable&)) 10)
                   (wlessp (wplus2 st 500) stkupbnd)) % stack available
             (setq oldstack st)
             (setq st (wplus2 st 257))   % 128 exceptions allowed by now
             (clear-memory oldstack 257) % clean it
             (if (setq heapupperbound (inspect-oldstable oldstack 256))
                        % heaplowerbound is set by insepct-oldstable
                 (setq small-variant t *small-garbage-collection* t)
                (progn
                 (setq st oldstack)
                 (setq *small-garbage-collection* NIL)
                 (setq heapupperbound oldheapupperbound
                       heaplowerbound oldheaplowerbound))))
 
      another-full-gc   % if heap is full
 
        (when *gc (errorprintf (if small-variant
                         "*** Small Vector Garbage collection starting"
                         "*** Vector Garbage collection starting")))
 
        (!%!&reclaim)      % the true gc
 
        (when (eq -1 (wdifference localhole heaplowerbound))
            (setq nothing-lower t))
        (when (eq -1 (wdifference heapupperbound localholeupper))
            (setq nothing-upper t))
        (setq heapupperbound oldheapupperbound
              heaplowerbound oldheaplowerbound )
 
        (rplaca &oldstable& (wdifference localhole heaplowerbound))
        (rplacd &oldstable& (wdifference heapupperbound
                                            (wplus2 2 localholeupper)))
        (when nothing-lower
             (setq localhole (wplus2 heaplowerbound -1))
             (rplaca &oldstable& (wdifference heaplast heaplowerbound)))
        (when nothing-upper
             (setq localholeupper (wplus2 heapupperbound 1))
             (rplacd &oldstable& (wplus2 -12
                            (wdifference heapupperbound heaptrapbound))))
 
        (when (and small-variant oldstack)
             (setq numberofexceptions (getmem oldstack))
             (when (and *gc *dump)
                    (putint numberofexceptions)
                    (console-print-string " Exceptions from stable"))
             (setq iii 1)
             (ifor (from i 1 numberofexceptions 1)
                  (do           % this is a gather operation
                   (setq adr (wgetv oldstack iii))
                   (putmem adr (wgetv oldstack (wplus2 iii 1)))
                   (setq iii (wplus2 iii 2))))
             (setq st oldstack))
 
        (postreclaim)
 
        (when *gc  (gcmessage))
 
        (when (or (wlessp (known-free-space) heap-warn-level)
                  (eq (wplus2 localheapshrinkupper
                              localheapshrinklower) 0))
               (when small-variant (setq small-variant nil
                                    *small-garbage-collection* NIL
                                    startingrealtime (timc))
                                   (go another-full-gc))
               (when (not (add-heap-space 10000))
                    (continuableerror 99 "Heap space low" nil)))
 
        (clear-frame)
        (when (wlessp (set-continue-at-attention nil) 0) (lisprpv))
 
)    )
 
(compiletime  % what to do with outside references from old stable part
 (ds damaged-stable (where item)
%   (putoct where ) (putoct item)
%    (putoct (wplus2 heap (car &oldstable&)))
%    (putoct (wdifference heapupperbound (cdr &oldstable&)))
%    (console-print-string "damaged")
    (if (wgeq exceptptr number-allowed)
       (progn (putmem except 0) (setq too-many-exceptions t))
       (progn (wputv except exceptptr where)
              (wputv except (wplus2 exceptptr 1) item)
              (putmem except (wplus2 (getmem except) 1))
              (setq exceptptr (wplus2 exceptptr 2)))
)) )
 
(de inspect-oldstable (except number-allowed)
 
% Idea: The stable part of the heap tends to remain stable for a
%       longer time and need not to be inspected very often. So
%       we try to raise heaplowerbound temporary, such that the
%       predicate withinheapointer will not be true for the old
%       stable part.
 
  (prog (currentitem ccont infccont endoldstable exceptptr
        too-many-exceptions endupperoldstable)
 
        (setq exceptptr 1)
        (setq endoldstable (wplus2 heaplowerbound (car &oldstable&)))
        (when (eq endoldstable heaplowerbound)
            (return    % nothing to do for lower heap
             (upper-inspect-oldstable except number-allowed exceptptr)))
 
        (setq endupperoldstable (wplus2 -1
                            (wdifference heapupperbound (cdr &oldstable&))))
        (setq currentitem  heaplowerbound )
        (setq ccont(getmem currentitem))
      loop
        (setq infccont (inf ccont))
        (setq currentitem (wplus2 currentitem 1))
        (case (tag ccont)
          (( btr-tag  code-tag negint-tag posint-tag id-tag unbound-tag ) NIL)
          (( string-tag bytes-tag fixnum-tag floatnum-tag bignum-tag
             words-tag pair-tag vector-tag evector-tag)
              (when (and (wgeq infccont endoldstable)
                         (wleq infccont endupperoldstable))
                          (damaged-stable (wplus2 currentitem -1) ccont)))
          (( hbytes-tag)
            (setq currentitem
                 (wplus2 currentitem (strpack (signedinf ccont)))))
          (( hwords-tag)
            (setq currentitem
                 (wplus2 currentitem (wrdpack (signedinf ccont)))))
          (( hvect-tag)
            (let (tmp tmp2)
              (setq tmp (signedinf ccont))
              (setq tmp2    (getmem currentitem))
              (ifor (from i 0 tmp 1) % VecLen + 1 items starting
               (do                   % behind the header word
                (let (tmp3 inftmp2)
                 (setq tmp3 (tag tmp2) inftmp2 (inf tmp2))
                 (when (and (wgeq inftmp2 endoldstable) (pointertagp tmp3)
                            (wleq inftmp2 endupperoldstable))
                              (damaged-stable currentitem tmp2))
                 (setq currentitem (wplus2 currentitem 1))
                 (setq tmp2    (getmem currentitem)))))))
           (nil
             (progn (center-dump currentitem 600)
              (gcerror " Error in stable scanning phase at %o" currentitem))))
 
          (setq ccont(getmem currentitem))
          (when (wlessp currentitem endoldstable) (go loop))
 
          (when too-many-exceptions (return nil))
          (setq heaplowerbound currentitem)
          (return
             (upper-inspect-oldstable except number-allowed exceptptr))
))  % if good up to this point, continue
 
(de upper-inspect-oldstable (except number-allowed exceptptr)
 
% Continuation of the inspection process for old stable part of heap
% Doing now the upper part of heap. First part has modified heaplower
% bound already if successful. The overall success depends on this
% second inspection.
 
(prog (addr length lowerendaddr upperendaddr too-many-exceptions
       position oldpos fourteen)
 
  (setq oldpos 0)
  (setq fourteen -14)
  (setq addr (aplus2 heapupperbound -1))
  (setq length (cdr &oldstable&))
  (setq lowerendaddr heaplowerbound)     %already raised by lower pass
  (setq upperendaddr (wplus2 1 (wdifference addr length)))
  (VsetVl 64)
  (vload (vreg 1) addr -1)                     % backwards
 
  (while (wgreaterp length 0)
    (VsetVL (if (wgreaterp length 64) 64 length))
    (setq oldpos -1)
 
    (prog (vmaskpointertag vmaskrange)
       (vrshiftv (vreg 0) (vreg 1) nontagpart)      % tags
                                                    % upper tag bound =15 
       (vplusS  (Vreg 6) (vreg 0) fourteen) 
                                                    % lower tag bound =0
       (setq vmaskpointertag  (vtestpos (Vreg 6)))  % tag >=codetag
       (minusv (Vreg 7) (vreg 0))
       (vlshiftv (vreg 3) (vreg 1) infstartingbit)  % inf-operation 1st part
       (setq vmaskpointertag (wor vmaskpointertag   % tag < code-tag
                                  (vtestpos (Vreg 7))))
 
    % vmaskpointer now indicates the vector elememts which are of type
    % pointer. Have to add information about pointing into nonstable part
 
       (vrshiftv (vreg 5) (vreg 3) infstartingbit) % inf-operation 2nd part
       (vsminusv (vreg 6) upperendaddr (vreg 5))   % upper bound
       (setq vmaskrange (vtestpos (vreg 6)))
       (vpluss (vreg 7) (vreg 5) (wminus lowerendaddr)) % lower bound
       (setq vmaskrange (wand vmaskrange (vtestpos (vreg 7))))
 
    % now its all together! Evaluate it.
 
       (setq vmaskrange (wand vmaskrange (wnot vmaskpointertag)))
     loop-over-exceptions
       (when (eq 0 vmaskrange) (return nil)) % ready, next vector block
 
    % exceptions found, isolate, make entry
 
       (setq position (wldzero vmaskrange))
         % accumulate the shifts
       (setq oldpos (wplus2 (wplus2 1 oldpos) position))
       (setq vmaskrange (wlshift vmaskrange (wplus2 position 1)))
                  % shift forward
       (damaged-stable (wdifference addr oldpos)
                       (getmem (wdifference addr oldpos)))
    % May also do a Vgetelement, but doesnt matter, rare case
       (go loop-over-exceptions)) %inner prog
 
    (setq addr   (aplus2 addr -64))
    (VsetVl 64)
    (vload (vreg 1) addr -1)                     % backwards
    (setq length (wdifference length 64))) % while
 
   (when too-many-exceptions (return nil))
   (return upperendaddr)
))
 
(de !%!&reclaim ()    %  The 'real' Garbage collector
     (prog ()
         (setq *gc-large-move NIL) % a critical move of heapspace
         (setq stackend  (wdifference st (framesize)))
         (setq stackstart stacklowerbound)
         (setf gcknt* (wplus2 gcknt* 1))   % gc counter
         (set-continue-at-attention t)     % dont worry about Attention
         (setq savereprieve *noreprieve)   % stack reprieve value
         (setq *noreprieve (when (not *dump) "garbage collector acting"))
         (setq heaptrapped t)
  (SETQ GCTIMING0 startingrealtime)
  (SETQ GCTIMING1 (timc))
         (markfromallbases)
  (SETQ GCTIMING2 (timc))
         (makeidfreelist)
  (SETQ GCTIMING3 (timc))
         (VsetVl 64)
         (vset (vreg 2) -1)  % Vreg 2 is used in relocation process
         (vstore (vreg 2) *gc-large-move-vector* 1)
         (buildrelocationfields)
  (SETQ GCTIMING4 (timc))
         (buildrelocationfieldsupper)
  (SETQ GCTIMING5 (timc))
         (updateallbases)
  (SETQ GCTIMING6 (timc))
         (compactheap)
  (SETQ GCTIMING7 (timc))
         (compactheapupper)
  (SETQ GCTIMING8 (timc))
         (setq heaplast (wdifference heaplast localheapshrinklower))
         (setq heaptrapbound (wplus2 heaptrapbound localheapshrinkupper))
 
         (setq *noreprieve savereprieve)
         (setq startingrealtime (wdifference (timc) startingrealtime))
         (setf gctime* (wplus2 gctime* startingrealtime))
         (setq heaptrapped nil)))
 
(compiletime      %macros for generation of pre- and post reclaim
 (progn           % save all fluids that have a *known-to-gc prop
 
  (fluid '(collectedregs*))
 
  (de collectregs-read (x)
    (when (and (get x 'registername) (get x '*known-to-gc))
     (setq collectedregs*
        (cons `(iputv gcextrasave ,(get x '*known-to-gc)
                                  ,x) collectedregs*))))
 
  (dm idtoreg-conversion-read ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-read))
       (cons 'progn collectedregs*) )
 
  (de collectregs-write (x)
    (when (and (get x 'registername) (get x '*known-to-gc))
     (setq collectedregs*
       (cons `(setf ,x (igetv gcextrasave ,(get x '*known-to-gc)))
                                                collectedregs*))))
 
  (dm idtoreg-conversion-write ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-write))
       (cons 'progn collectedregs*) )
))
 
(de prereclaim ()     % save critical from register to heap
    (idtoreg-conversion-read))
 
(de postreclaim ()     % restore critical from heap to register
    (idtoreg-conversion-write))
 
(fluid '(*gc-large-move-vector*))
 
(setq *gc-large-move-vector* (gtwarray 65))  % for large heaps > 4 mio
 
(de searchvec (y)   % match interval in vector register (Vreg 2)
  (prog (mask)
    (VsetVl 64)                      % could  be  *gc-large-move
    (VandS (Vreg 3) 16#ffffffff (vreg 2))
    (VSminusV (Vreg 1) y (Vreg 3))
    (setq mask (wnot (VtestNEG (Vreg 1)) ))
    (when (eq 1 mask) (return 0))
    (setq y (vgetelement (Vreg 2) (wldzero mask)))
    (return (wrshift y 32))))
 
(de markfromallbases ()
  (prog (b)
      (setq localhole (wplus2 heapupperbound -2))
           % multiple use of localhole for marking phase
           (setq schaedling 1)
      (markfromsymbols)
           (setq schaedling 2)
      (markfromrange stackstart stackend)
           (setq schaedling 3)
      (markfromvectorregisters)
      (setq b bndstklowerbound)
           (setq schaedling 4)
      (while (progn (setq b (adjustbndstkptr b 1))
                      (wleq b bndstkptr))
             (markfrombase (getmem b)))))
 
(de markfromvectorregisters () nil)  % to be replaced later
% two little helper for gc
 
(de last-string-in-symnam ()
 
  % helps to find last occupied id slot (a string) in symnam vector
  % speeding up garbage collector a little bit. May NEVER user Vreg 2!!
 
    (prog (endaddr mask)
      (setq endaddr (wplus2 symnam maxsymbols))
 
      (while (wgreaterp endaddr symnam) % loop doubled
         (VsetVl 64)
          % we must not care about exact Vector length since we cannot
          % reach the end .low ids are always marked!
         (vload (vreg 1) endaddr -1)       % backwards
         (vload (vreg 6) (wplus2 endaddr -64) -1)       % backwards
         (vrshiftv (vreg 5) (vreg 1) nontagpart)
         (Vsminusv  (vreg 3) string-tag (vreg 5))
         (setq mask (vtestn (vreg 3)))
         (if (eq mask -1) (setq endaddr (wplus2 endaddr -64))
                          (return NIL))
         (vrshiftv (vreg 5) (vreg 6) nontagpart)
         (Vsminusv  (vreg 3) string-tag (vreg 5))
         (setq mask (vtestn (vreg 3)))
         (if (eq mask -1) (setq endaddr (wplus2 endaddr -64))
                          (return NIL)))
 
      (return (wdifference endaddr symnam))))
 
(de last-noninteger-in-symnam ()
 
  % helps to find last noninteger slot in symnam vector
  % This is a freed id. After that no update to symnam necessary.
 
    (prog (endaddr mask)
      (setq endaddr (wplus2 symnam maxsymbols))
 
      (while (wgreaterp endaddr symnam) % loop doubled
         (VsetVl 64)
          % we must not care about exact Vector length since we cannot
          % reach the end .low ids are always marked!
         (vload (vreg 1) endaddr -1)       % backwards
         (vload (vreg 6) (wplus2 endaddr -64) -1)       % backwards
         (vrshiftv (vreg 5) (vreg 1) nontagpart)
         (Vsminusv  (vreg 3) posint-tag (vreg 5))
         (setq mask (vtestz (vreg 3)))
         (if (eq mask -1) (setq endaddr (wplus2 endaddr -64))
                          (return NIL))
         (vrshiftv (vreg 5) (vreg 6) nontagpart)
         (Vsminusv  (vreg 3) posint-tag (vreg 5))
         (setq mask (vtestz (vreg 3)))
         (if (eq mask -1) (setq endaddr (wplus2 endaddr -64))
                          (return NIL)))
      (return (if (wlessp (setq mask (wdifference endaddr symnam)) maxsymbols)
                  (wplus2 mask 3) maxsymbols))))
 
(de markfromsymbols()
 (if *small-garbage-collection* % mark ALL symbols, because ids in stable
  (prog (b b1 b2 b3 rem j k l maxi)  % part are not reported in marking process
    (setq maxi (last-string-in-symnam))
    (markfromonesymbol 128 (symnam 128)) % mark NIL first
    (ifor (from i 0 127 1) (do (markfromonesymbol i (symnam i))))
    (setq rem (wdifference maxi (wlshift (wrshift maxi 2) 2)))
    (ifor (from i 0 (wplus2 rem -1) 1) % rest loop
            (do (progn (setq b (symnam i))
                       (when (stringp b) (markfromonesymbol i b)))))
    (ifor (from i rem maxi 4)
            (do (progn (setq b  (symnam i)) (setq j (wplus2 i 1))
                       (setq b1 (symnam j)) (setq k (wplus2 j 1))
                       (setq b2 (symnam k)) (setq l (wplus2 k 1))
                       (setq b3 (symnam l))
                       (when (stringp b)  (markfromonesymbol i b))
                       (when (stringp b1) (markfromonesymbol j b1))
                       (when (stringp b2) (markfromonesymbol k b2))
                       (when (stringp b3) (markfromonesymbol l b3))))))
  (prog (b  b1 b2 b3 rem)
    (markfromonesymbol 128 (symnam 128)) % Mark NIL first
    (ifor (from i 0 127 1) (do (markfromonesymbol i (symnam i))))
 
    (setq rem  (wlshift (wrshift hash-table-size 2) 2))
    (ifor (from i 0 rem 4)
      (do (progn          % taking 4 cells in parallel
           (setq b (getmem (aplus2 hashtable (wrshift i 2))))
           (setq b1 (wand (wrshift b 32) 16#FFFF))
           (setq b2 (wand (wrshift b 16) 16#FFFF))
           (setq b3 (wand b 16#FFFF))
           (setq b  (wrshift b 48))
           (when (wlessp b1 16#ffff) (markfromonesymbol b1 (symnam b1)))
           (when (wlessp b2 16#ffff) (markfromonesymbol b2 (symnam b2)))
           (when (wlessp b3 16#ffff) (markfromonesymbol b3 (symnam b3)))
           (when (wlessp b  16#ffff) (markfromonesymbol b  (symnam b ))))))
    (ifor (from i (wplus2 rem 1) hash-table-size 1) % rest loop
        (do (progn (setq b (hash-table-entry i))
               (when (wgreaterp b 0) (markfromonesymbol b (symnam b)))))))))
 
(de markfromonesymbol (b symn)
% SymNam has to be marked from before marking ID, since the mark uses its tag
% No problem since it's only a string, can't reference itself.
% gets symnam entry with the call.
       (setq schaedling1 b)
       (when (markedidinSymnam symn) (return b))
       (setq schaedling2 0)
       (markfrombase symn)
       (setq schaedling2 1)
       (markidinSymnam b symn)
       (setq schaedling2 2)
       (markfrombase (symprp b))
       (setq schaedling2 3)
       (markfrombase (symget b))
       (setq schaedling2 4)
       (markfrombase (symval b))
       (setq schaedling2 5)   )
 
(de markfromonesymbol-1 (b symn)
% SymNam has to be marked from before marking ID, since the mark uses its tag
% No problem since it's only a string, can't reference itself.
% gets symnam entry with the call.
       (setq schaedling1a b)
       (when (markedidinSymnam symn) (return b))
       (setq schaedling2a 0)
       (markfrombase symn)
       (setq schaedling2a 1)
       (markidinSymnam b symn)
       (setq schaedling2a 2)
       (markfrombase (symprp b))
       (setq schaedling2a 3)
       (markfrombase (symget b))
       (setq schaedling2a 4)
       (markfrombase (symval b))
       (setq schaedling2a 5)   )
 
(de markfrombase (base)
    (prog (markadr markinf marktg)
         (setq marktg (tag base))
         (when (null base) (return base))
         (case marktg
           ((posint-tag negint-tag code-tag btr-tag unbound-tag)
              (return NIL))
           ((id-tag)
              (progn
               (setq markadr (inf base))
              (return (markfromonesymbol-1 markadr (symnam markadr)))))
           ((fixnum-tag bignum-tag floatnum-tag string-tag
             bytes-tag words-tag)
              (progn
               (setq markadr (inf base))
               (when (not (withinheappointer markadr)) (return nil))
               (setq markinf (getmem markadr))
                  (when (and (wlessp markadr heaptrapbound)
                             (posintp markinf)
                             (posintp (wgetv markadr 1)))
                     (console-print-string "Totschlag")
                     (backtrace) (putheapdata base)
                     (center-dump base 300)
                     (center-dump markadr 300)
                     (print (int2id schaedling1))
                     (print (symval (int2id schaedling1)))
                     (print (symnam (int2id schaedling1)))
                     (print (symprp (int2id schaedling1)))
                     (print (symfnc (int2id schaedling1)))
                     (print (symget (int2id schaedling1)))
                     (terpri)
                     (print (int2id schaedling1a))
                     (print (symval (int2id schaedling1a)))
                     (print (symnam (int2id schaedling1a)))
                     (print (symprp (int2id schaedling1a)))
                     (print (symfnc (int2id schaedling1a)))
                     (print (symget (int2id schaedling1a)))
                     (print (symval (int2id 128)))
                     (print (symnam (int2id 128)))
                     (print (symprp (int2id 128)))
                     (print (symfnc (int2id 128)))
                     (print (symget (int2id 128))))
               (putmem markadr (insmark markinf gcmarkvalue))
               (return NIL)))
           ((pair-tag)
              (progn
               (setq markadr (inf base))
               (when (not (withinheappointer markadr)) (return nil))
               (setq markinf (getmem markadr))
               (when (weq (gcfield markinf) gcmarkvalue) (return nil))
                  (when (and (wlessp markadr heaptrapbound)
                             (posintp markinf)
                             (posintp (wgetv markadr 1)))
                     (console-print-string "MORD")
                     (backtrace) (putheapdata base)
                     (center-dump base 300)
                     (center-dump markadr 300))
               (putmem markadr (insmark markinf gcmarkvalue))
               (return (markfrombaseaux markinf (cdr base)))))
           ((vector-tag evector-tag)
              (progn
               (setq markadr (inf base))
               (when (not (withinheappointer markadr)) (return nil))
               (setq markinf (getmem markadr))
               (when (weq (gcfield markinf) gcmarkvalue) (return nil))
                  (when (and (wlessp markadr heaptrapbound)
                             (posintp markinf)
                             (posintp (wgetv markadr 1)))
                     (console-print-string "Gemeiner Totschlag")
                     (backtrace) (putheapdata base)
                     (center-dump base 300)
                     (center-dump markadr 300))
               (putmem markadr (insmark markinf gcmarkvalue))
               (return (markfromvector markadr))))
           ((hbytes-tag halfwords-tag hwords-tag hvect-tag forward-tag
                        hhalfwords-tag)
                % those are ignored
               (return  NIL))
           (nil (progn (putheapdata base)
                       (console-print-string " ******* Illegal item in mark")
                       (putint marktg) (console-newline)
                     (return NIL))))
))
 
 
% One of the workhorses of garbage collector:
% Marking pairs. Alloc is necessary only if cdr is non-NIL.
% Special coded stack management (experimental)
 
(if_system Cray-2
(lap '((*entry markfrombaseaux expr 2)
       (*move (quote NIL) (reg s7))
       (*wxor3 (Reg s0) (Reg 2) (reg s7))
% (JZ (reg s0) (PLUS2 (parcel symfnc) (times2 4 (idloc markfrombase))))
       (Jz (reg s0) (internalentry markfrombase))
 
       (*alloc 1)
       (*move (reg 2) (frame 1))
       (*call markfrombase)
       (*move (frame 1) (reg 1))
       (*dealloc 1)
       (*jcall markfrombase)))
 
(lap '((*entry markfrombaseaux expr 2) % Cray1,X-MP,Y-MP
       (*move (quote NIL) (reg s7))
       (*wxor3 (Reg s0) (Reg 2) (reg s7))
    %% (JSZ (PLUS2 (parcel symfnc) (times2 4 (idloc markfrombase))))
       (Jsz (internalentry markfrombase))
 
       (*alloc 1)
       (*move (reg 2) (frame 1))
       (*call markfrombase)
       (*move (frame 1) (reg 1))
       (*dealloc 1)
       (*jcall markfrombase))))
 
(de markfromrange (low high)
  (ifor (from ptr low high 1) (do (markfrombase (getmem ptr)))))
 
(de externalmarkfrombase (base)  % for external reference
   (markfrombase  base))
 
(de markfromvector (info)
  (prog (uplim)
        (commentoutcode (when (wneq (tag (getmem info))  hvect-tag)
                          (return nil)))
        (setq uplim (loc (vecitm info (veclen info))))
        (ifor (from ptr (loc (vecitm info 0)) uplim 1)
              (do (progn (markfrombase (getmem ptr))
) )     )     )   )
 
(de makeidfreelist ()
  (prog (previous maxi)
        (setq maxi (last-noninteger-in-symnam))
        (ifor (from i 0 128 1) (do (clearidmark i)))
        (setq previous 129)
        (while (and (markedid previous) (wleq previous  maxsymbols))
             (clearidmark previous)
             (setq previous (wplus2 previous 1)))
        (if (wgeq previous maxsymbols)
             (setq nextsymbol 0)
             (setq nextsymbol previous))
        (ifor % free list starts here
              (from i (wplus2 previous 1) maxi 1)
              (do (if (markedid i)
                    (clearidmark i)
                    (progn (setf (symnam previous) i)
                           (setq previous i)))))
        (if (wlessp maxi maxsymbols)
                    (setf (symnam (wplus2 maxsymbols -1)) 0)
                    (setf (symnam previous) 0))))
 
 
(de buildrelocationfieldsupper ()
 
% Looking for NON marked pairs in the upper heap.
% Clearing mark bits on the fly. Special caution for negative integers
 
(prog (addr length lowerendaddr upperendaddr too-many-exceptions
       notgcfieldmask
       gcfieldmask shiftedgcmarkval nn vmasknegint vmaskmark)
 
  (setq gcfieldmask (wrshift -1 (plus2 infbitlength tagbitlength)))
  (setq localholeupper (wplus2 heapupperbound 1))
  (setq gcfieldmask (wlshift gcfieldmask infbitlength))
  (setq notgcfieldmask (wnot gcfieldmask))
  (setq shiftedgcmarkval (wlshift gcmarkvalue infbitlength))
 
  (setq addr (wplus2 heapupperbound -2))
  (setq length (wquotient (wdifference addr (wplus2 heaptrapbound 10)) 2))
  (VsetVl 64)
  (vload (vreg 1) addr -2)                     % backwards,pairwise
 
  (while (wgreaterp length 0)
       (VsetVL (if (wgreaterp length 64) 64 length))
 
       (vrshiftv (vreg 0) (vreg 1) nontagpart)      % tag
       (vsminusV (Vreg 3) negint-tag (vreg 0))
       (setq vmasknegint (vtestz (Vreg 3)))         % tag = negint-tag
       (Vands (vreg 4) gcfieldmask (vreg 1))
       (vsminusV (Vreg 2) shiftedgcmarkval (vreg 4))
       (setq vmaskmark (vtestn (Vreg 2)))
 
    % vmaskmark should be all zero. If not we found a nonmarked pair
    % Note the holeaddr in lcalholeupper and set Vl for clearing
    % accordingly
 
       (when (not (eq vmaskmark 0))
          (setq nn (wldzero vmaskmark))
          (setq localholeupper (wdifference addr (wtimes2 nn 2)))
          (setq length 0)  % stop loop
          (VsetVl (wplus2 nn 1)))
 
       (vands (vreg 2) notgcfieldmask (Vreg 1))  % cleared values
       (vors  (vreg 3) gcfieldmask (vreg 2))     % for negints
       (vsetvm vmasknegint)
       (vmergeVv (vreg 4) (vreg 3) (vreg 2))
       (vstore (vreg 4) addr -2)   % update cleared vector
 
       (setq addr (wdifference addr 128))
       (vload (vreg 1) addr -2)                     % backwards,pairwise
       (setq length (wdifference length 64))
  ) % end while
  (when nn (return (buildrelocationfieldsupper1 localholeupper)))
  (setq localheapshrinkupper 0)
))
 
(compiletime (progn
 
(ds addzeros (n)
   (prog (ntimes2)
     (setq ntimes2 (wlshift n 1))
     (setq holelength (wplus2 holelength ntimes2)) % items
     (setq dcount (wplus2 dcount ntimes2))
     (setq heapshr (wplus2 heapshr ntimes2))
     (when (eq holelength ntimes2)         % hole just started
                   (setq holestart pp))
     (setq pp (wdifference pp ntimes2))    % current vectaddr
     (when (eq status 1)                   % hole finished
         (putmem holestart (mkitem hskip holelength)))
)  )
 
(ds buildreloc (n)
   (prog (ntimes2 nnnn cont target target1)
     (setq cont (getmem pp))
     (setq nnnn n)
     (setq ntimes2 (wlshift nnnn 1))
     (setq holelength 0)
     (when (eq n 0) (return n))
 
    buildloop            % a simple one first attempt
 
     (when (wneq (gcfield cont) 511)
           (putheapdata pp)
           (center-dump pp 64)
           (console-print-string " ueberreagiert"))
 
       % The absolute target location for the item will be placed
       % in the gc field. So targets above 4 M words cannot be handled
       % easily. We maintain a relocation vector for this purpose, which
       % defines a mapping between currentitem and currentoffset.
 
     (setq target (wplus2 pp dcount))
  test_offset
     (setq target1 (wdifference target current_offset))
     (when (wlessp target1 0)       % overflow of current_offset
            (if *gc-large-move  % already active
                 (progn
                   (setq *gc-large-move (wplus2 *gc-large-move 1))
                   (when (wgreaterp *gc-large-move 64)
                          (throw '*gc-large-move pp)))
                           % let the World Know that something happened
                           % (a move > 256 MW !) Nobody will catch it.
                 (setq *gc-large-move 0))  %vectorposition
 
            (wputv *gc-large-move-vector* *gc-large-move
                         (wor (wlshift current_offset 32)
                              (wand (wplus2 pp 1) 16#ffffffff)))
            (setq current_offset (wdifference current_offset criticaloffset))
 
          % put into vector
            (go test_offset))
 
     (putmem pp (insmark cont target))%%target1))
     (setq pp (wplus2 pp -2))
     (setq cont (getmem pp))
     (when (wgreaterp (setq nnnn (wplus2 nnnn -1)) 0) (go buildloop))
)  )
)  )
 
(de buildrelocationfieldsupper1 (currentitem)
 
(prog (length status veclength mask notmask gcfieldmask current_offset
       criticaloffset nn pp heapshr dcount holestart holelength)
 
  (setq criticaloffset critical-offset) % may be a non immediate value
  (setq current_offset
     (wlshift (wrshift currentitem critical-offset-bitsize)
                                   critical-offset-bitsize))
  (setq *gc-large-move -1)
  (setq localheapshrinkupper 2 heapshr 2 dcount 2 holelength 2)
  (setq holestart currentitem)
  (setq gcfieldmask (wrshift -1 (plus2 infbitlength tagbitlength)))
  (setq currentitem (aplus2 currentitem -2))
  (setq status 0)   % in a hole
  (setq length (wquotient  % number of pairs
                 (wdifference currentitem (wplus2 10 heaptrapbound)) 2))
 
  (vsetvl 64)
  (vset (vreg 6) gcfieldmask)  % presetting vector registers with useful
  (vset (vreg 7) gcmarkvalue)  % values . CAUTION do not alter
  (vload (vreg 1) currentitem -2)  % loading backwards, pairwise
 
  (while (wgreaterp length 0)
 
    (VsetVL (setq veclength (if (wgreaterp length 64) 64 length)))
 
    (setq pp currentitem)
    (vrshiftv (vreg 0) (vreg 1) infbitlength)    % inf part
    (vandV (vreg 5) (vreg 0) (vreg 6))           % isolate gcfield
    (vminusv (Vreg 3) (vreg 5) (vreg 7))
    (setq mask (vtestz (vreg 3)))
    (setq notmask (wnot mask))
 
  % we have obtained a mask indicating all marked pairs by a 1 bit and
  % a second one which we use for length calculation
 
  % first two important special cases:
 
    (when (and (eq mask 0) (eq status 0)) % all pairs NOT marked
       (addzeros veclength) (go bottom))    % continuation of a hole
 
    (when (and (eq mask -1) (eq status 1)) % all pairs MARKED
       (buildreloc veclength) (go bottom))   % and last pair marked too
 
    Mixed-case
 
    (when (eq status 0)              % we are in a hole
 
      (setq nn (wldzero mask))       % how many zeros following
      (when (eq nn 64) (addzeros veclength) (go bottom))
                                     % no more marked items
      (setq mask (wlshift mask nn))
      (setq notmask (wlshift notmask nn)) % hole is finished
      (setq status 1)
      (addzeros nn)
      (setq veclength (wdifference veclength nn))) %eaten
 
    (when (and (wgeq veclength 0) (eq status 1)) % we are on the green
 
      (setq nn (wldzero notmask))    % how many ones following
      (when (eq nn 64) (buildreloc veclength) (go bottom))
                                     % buildreloc for the rest
      (setq mask (wlshift mask nn))
      (setq notmask (wlshift notmask nn)) % off the green
      (setq status 0)
      (buildreloc nn)
      (setq veclength (wdifference veclength nn))) %eaten
    (when (wgreaterp veclength 0) (go mixed-case))
   bottom
    (VsetVl 64)
    (setq currentitem (wplus2 currentitem -128))
    (vload (vreg 1) currentitem -2)    % loading backwards, pairwise
    (setq length (wplus2 length -64)) )  % end of while loop
  (setq status 1)
  (addzeros 0)  % write last hskip if in a hole
  (setq localheapshrinkupper heapshr)
 
  % cleaning the relocation vector
  (setq currentitem (wplus2 heaptrapbound 6))
  (ifor (from i (wplus2 1 *gc-large-move) 62 1) (do
        (when (eq -1 (wgetv *gc-large-move-vector* i))
          (wputv *gc-large-move-vector* i
                 (wor (wlshift current_offset 32)
                 (wand currentitem 16#ffffffff))))))
  (wputv *gc-large-move-vector* 63 0)
               % We have to fill the largemove vector
))
 
(de buildrelocationfields ()
  %
  %  Start of Pass 2 - Turn off GC marks and Build SEGKNTs
  %
  (prog (currentitem dcount seglen holestart holelength
         currentcont indic clearedvalue mask1 mask2)
 
        (setq mask1 (wlshift 8#17777777 infbitlength))
        (setq mask2 (wnot mask1))
        (setq localhole (wdifference heaplowerbound 1))
        (setq holelength (setq dcount (setq localheapshrinklower 0)))
        (setq currentitem  heaplowerbound )
        (setq currentcont (getmem currentitem))
        (when (wgeq currentitem heaplast) (return nil))
                  % may be empty at all
  loop
        (setq seglen 2)  % most cases
        (setq indic (wdifference (gcfield currentcont) gcmarkvalue))
        (setq clearedvalue (wand currentcont mask2))
        (case (tag currentcont)
           ((negint-tag) (setq clearedvalue (wor currentcont mask1)))
           ((hbytes-tag)   %  h-tags  must be first of pair
              (setq seglen (wplus2 1 (strpack (signedinf currentcont)))))
           ((hwords-tag)
              (setq seglen (wplus2 1 (wrdpack (signedinf currentcont)))))
           ((hvect-tag)
              (setq seglen (wplus2 1 (vectpack (signedinf currentcont)))))
           ((hskip) (setq seglen (inf currentcont)))
           ((btr-tag code-tag evector-tag posint-tag id-tag unbound-tag
             string-tag bytes-tag fixnum-tag floatnum-tag bignum-tag
             words-tag halfwords-tag pair-tag vector-tag) NIL)
           (nil
             (progn (center-dump currentitem 600)
                    (gcerror "Illegal item in heap at %o" currentitem))))
 
       (when (wneq indic 0) % first hole in heap , change to  ..1
           (setq localhole currentitem)
           (return (buildrelocationfields1 currentitem seglen)))
       (putmem currentitem clearedvalue)
       (setq currentitem (wplus2 currentitem seglen))
 
       (setq currentcont (getmem currentitem))
       (when (wlessp currentitem heaplast) (go loop))
       (when (wgreaterp currentitem heaplast) (putheapdata currentitem))
 ))
 
(de buildrelocationfields1 (currentitem seglen)
  %
  %  continuation of Pass 2 - Turn off GC marks and Build SEGKNTs
  %
  % currentitem is not marked, we will start a hole:
  (prog (localdust dcount holestart holelength indic currentcont
         current_offset rel_offset heapshr target criticaloffset)
 
     (setq criticaloffset critical-offset)  % may be no immediate value
     (setq *gc-large-move NIL)
     (when (wgreaterp currentitem heaplast) (putheapdata currentitem))
     (setq current_offset 0 rel_offset 0)
     (setq localheapshrinklower (setq heapshr seglen))
     (setq dcount seglen)
     (setq holelength seglen)
     (setq holestart currentitem)
     (setq currentitem (wplus2 currentitem seglen))
     (setq localdust TREGSAVE)      % since TREGSAVE is temp storage
     (setq currentcont (getmem currentitem))
     (when (wgeq currentitem heaplast) (go special-exit))
  loop
     (setq seglen 2)  % most cases
     (setq indic (wdifference (gcfield currentcont) gcmarkvalue))
     (when (wlessp (tag currentcont) 10) (go simple-case))
     (case (tag currentcont)
       ((hbytes-tag)     % must be first of pair
           (setq seglen (wplus2 1 (strpack (signedinf currentcont)))))
       ((hwords-tag)
           (setq seglen (wplus2 1 (wrdpack (signedinf currentcont)))))
       ((hvect-tag)
           (setq seglen (wplus2 1 (vectpack (signedinf currentcont)))))
       ((hskip)  (setq seglen (inf currentcont)))
       ((btr-tag code-tag evector-tag posint-tag negint-tag
         id-tag unbound-tag string-tag bytes-tag fixnum-tag
         floatnum-tag bignum-tag words-tag halfwords-tag
         pair-tag vector-tag) NIL)
       (nil
         (progn (center-dump currentitem 600)
                (gcerror "Illegal item in heap at %o" currentitem))))
 
  simple-case
     (if (weq indic 0) % marked item
        (prog(target1)
 
       % The absolute target location for the item will be placed
       % in the gc field. So targets above 4 M words cannot be handled
       % easily. We maintain a relocation vector for this purpose, which
       % defines a mapping between currentitem and currentoffset.
 
           (setq target (wdifference currentitem dcount))
        test_offset
           (setq target1 (wdifference target current_offset))
           (when (wneq (wrshift target1 critical-offset-bitsize) 0)
                                     % overflow of current_offset
              (if *gc-large-move  % already active
                  (progn
                   (setq *gc-large-move (wplus2 *gc-large-move -1))
                   (when (wlessp *gc-large-move 1)
                          (throw '*gc-large-move currentitem)))
                           % let the World Know that something happened
                           % (a move > 256 MW !) Nobody will catch it.
                  (setq *gc-large-move 62))  %vectorposition
 
              (setq current_offset (wplus2 current_offset criticaloffset))
              (wputv *gc-large-move-vector* *gc-large-move
                     (wor (wlshift current_offset 32)
                          (wand currentitem 16#ffffffff)))
          % put into vector
              (go test_offset))
 
           (putmem currentitem    % ABSOLUTE target location
                   (insmark currentcont target))%target1))
           (putmem holestart (mkitem hskip holelength))
           (setq holelength 0)
           (setq holestart localdust))  % put into dust
                   %not marked
        (setq heapshr (wplus2 heapshr seglen))
        (setq dcount (wplus2 dcount seglen))
 
        (setq holelength (wplus2 holelength seglen)) % this includes =0 case
        (when (weq holelength seglen)  % start of hole
                  (setq holestart currentitem)))
 
    (setq currentitem (wplus2 currentitem seglen))
    (setq currentcont(getmem currentitem))
    (when (wlessp currentitem heaplast) (go loop))
 
    (when (and *dump (wgreaterp currentitem heaplast))
                        (putheapdata currentitem)) %test
  special-exit % Only One hole at the very end
    (when (not (weq holelength 0))
           (putmem holestart (mkitem hskip holelength)))
    (setq localheapshrinklower heapshr)
    (if *gc-large-move
      (wputv *gc-large-move-vector* (wplus2 *gc-large-move -1)
                        (wor (wlshift current_offset 32)
                             (wand currentitem 16#ffffffff)))
      (wputv *gc-large-move-vector* 61
                        (wor (wlshift current_offset 32)
                             (wand currentitem 16#ffffffff))))
 ))
 
 
(compiletime
(ds updateitem (x)
  ((lambda (ptr)
     (prog (tg info cont)
        (setq tg (tag (setq cont (getmem ptr))))
        (unless (pointertagp tg) (return nil))
        (setq info (inf cont))
        (when (not (within-nonstable-heap info)) (return nil))
        (putmem ptr (mkitem tg (reloc info)))
   )) x))  )
 
(de updateallbases ()
  (prog (b oldlocalhole oldlocalholeupper)
  %
  % Have to care about the case when nothing is freed in either
  % part of the heap. Modifying localhole pointers accordingly
  %
        (setq oldlocalhole localhole)
        (setq oldlocalholeupper localholeupper)
        (when (wlessp localhole heaplowerbound)
              (setq localhole heaplast))
        (when (wgreaterp localholeupper heapupperbound)
              (setq localholeupper (wplus2 heaptrapbound 8)))
        (VsetVl 64)
        (vload (vreg 2) *gc-large-move-vector* 1)
                          % this vector must be loaded now!
        (updatesymbols)
        (updateregion stackstart stackend)
        (updatevectorregisters)
        (setq b bndstklowerbound)
        (while (progn (setq b (adjustbndstkptr b 1)) (wleq b bndstkptr))
                (updateitem b))
        (updateheap)
        (updateheapupper)
        (setq localhole      oldlocalhole)
        (setq localholeupper oldlocalholeupper) ))
 
(de updatesymbols ()
  (prog (nameloc maxi)
     (setq maxi (last-string-in-symnam))   % actually occupied id slots
     (ifor (from i 0 maxi 1)
                (do
                  (setq nameloc (loc (symnam i)))
                  (when (stringp (getmem nameloc))
                    (setq schaedling i)
                    (updateitem nameloc)
                    (updateitem (loc (symval i)))
                    (updateitem (loc (symget i)))
                    (updateitem (loc (symprp i))))))))
 
(de updateregion (low high)
  (setq schaedling 'region)
  (ifor (from ptr low high 1) (do (setq schaedling1 ptr)
                                  (updateitem ptr))))
%%(ifor (from ptr low high 1) (do (updateitem ptr))))
 
(de updatevectorregisters () nil)  % to be replaced later
 
(de updateheap ()
  (setq schaedling 'heap)
  (prog (currentitem ccont infccont)
        (setq currentitem  heaplowerbound )
        (when (wgeq currentitem heaplast) (return nil))
      loop
        (setq ccont(getmem currentitem))
        (setq infccont (inf ccont))
        (setq schaedling1 infccont)
        (setq currentitem (wplus2 currentitem 1))
        (prog nil
          (case (tag ccont)
            ((btr-tag code-tag negint-tag posint-tag id-tag
              unbound-tag ) NIL)
            ((string-tag bytes-tag fixnum-tag floatnum-tag bignum-tag
              words-tag halfwords-tag pair-tag vector-tag evector-tag)
               (progn (when (within-nonstable-heap infccont)
                          (putmem (aplus2 currentitem -1)
                                 (insinf ccont (reloc infccont))))))
             ((hbytes-tag)
                 (setq currentitem
                    (wplus2 currentitem (strpack (signedinf ccont)))))
             ((hwords-tag)
                  (setq currentitem
                     (wplus2 currentitem (wrdpack (signedinf ccont)))))
             ((hvect-tag)
                      (prog (tmp)
                        (setq tmp (signedinf ccont))
                           % Move over header
                        (ifor (from i 0 tmp 1) (do % VecLen + 1 items
 
                            (prog (tmp2 tmp3 inftmp2)
                                   (setq tmp2 (getmem currentitem))
                                   (setq tmp3 (tag tmp2))
                                   (setq inftmp2 (inf tmp2))
                                   (when (and (pointertagp tmp3)
                                        (within-nonstable-heap inftmp2))
                                      (setf (inf (getmem currentitem))
                                                 (reloc inftmp2)))
                                    (setq currentitem
                                      (wplus2 currentitem 1)))))))
             ((hskip)
                 (setq currentitem
                    (wplus2 -1 (wplus2 currentitem (signedinf ccont)))))
             (nil
                (progn
                 (center-dump currentitem 600)
                 (gcerror "Internal error in updating phase at %o"
                    currentitem)))))
          (when (wlessp currentitem heaplast) (go loop))
          (when (wgreaterp currentitem heaplast)
                               (putheapdata currentitem)) %test
                              ))
 
(de updateheapupper ()
 %
 % In the upper part of heap we expect to find pairs or skip marks
 % the later on car position only. So if we have no hskip, we can update
 % the next two entries.
 %
  (prog (currentitem ccont infccont currentitem2 ccont2 infccont2
         rangeend)
     (setq rangeend (wplus2 heaptrapbound 10))
     (setq currentitem (wplus2 heapupperbound -2))
     (go loop-bottom)
   loop
     (setq infccont  (inf ccont))
     (setq infccont2 (inf ccont2))
     (when (eq (tag ccont) hskip)  % test skip tag first
         (setq currentitem (wdifference currentitem (inf ccont)))
         (go loop-bottom))
     (when (and (pointertagp (tag ccont)) % else update a pair
                (within-nonstable-heap infccont))
               (putmem currentitem (insinf ccont (reloc infccont))))
     (when (and (pointertagp (tag ccont2)) % else update a pair
                (within-nonstable-heap infccont2))
               (putmem (aplus2 currentitem 1)
                            (insinf ccont2 (reloc infccont2))))
     (setq currentitem (wplus2 currentitem -2))
   loop-bottom
     (setq ccont(getmem currentitem))
     (setq ccont2 (getmem (aplus2 currentitem 1)))
     (when (wgreaterp currentitem rangeend) (go loop))
))
 
(de externalupdateitem (base)  % for external reference
   (updateitem  base))
 
(de compactheap ()
  (prog (olditemptr newitemptr seglen oldcont clearedvalue mask1 mask2)
        (setq mask1 (wlshift 8#17777777 infbitlength))
        (setq mask2 (wnot mask1))
        (when (wlessp localhole  heaplowerbound) (return nil))
        (setq newitemptr (setq olditemptr localhole))
        (while (wlessp olditemptr heaplast)
          (prog nil
             (setq seglen 2)
             (setq oldcont (getmem olditemptr))
             (setq clearedvalue (wand oldcont mask2))
             (case (tag oldcont)
                ((negint-tag) (setq clearedvalue (wor oldcont mask1)))
                ((btr-tag code-tag posint-tag id-tag unbound-tag
                  string-tag bytes-tag fixnum-tag floatnum-tag pair-tag
                  bignum-tag words-tag halfwords-tag vector-tag
                  evector-tag) NIL)
                ((hbytes-tag)
                   (setq seglen
                     (wplus2 1 (strpack (signedinf oldcont)))))
                ((hwords-tag)
                   (setq seglen
                     (wplus2 1 (wrdpack (signedinf oldcont)))))
                ((hvect-tag)
                    (setq seglen
                     (wplus2 1 (vectpack (signedinf oldcont)))))
                ((hskip)
                   (progn
                    (setq olditemptr (wplus2 olditemptr (inf oldcont)))
                    (go whilenext)))
                (nil
                    (progn
                    (center-dump olditemptr 600)
                    (gcerror "Internal error in compaction at %o"
                     olditemptr)))  )
             % case
           (putmem olditemptr clearedvalue)
           (VSetVl seglen)
           (Vload (Vreg 1) olditemptr 1)
           (if (wgreaterp seglen 64)
                 (vector-copy olditemptr newitemptr seglen)
                 (Vstore (Vreg 1) newitemptr 1))
 
          (setq newitemptr (wplus2 newitemptr seglen))
          (setq olditemptr (wplus2 olditemptr seglen))
      whilenext
        nil))
))
 
(de compactheapupper ()
 % compact heap in upper part (pairs only)
 % dont expect high tags etc
  (prog (olditemptr newitemptr oldcont clearedvalue rangeend
         mask1 mask2 cdrval)
        (setq mask1 (wlshift 8#17777777 infbitlength))
        (setq mask2 (wnot mask1))
        (setq rangeend (wplus2 heaptrapbound 10))
        (when (wgreaterp localholeupper heapupperbound) (return nil))
        (setq newitemptr (setq olditemptr localholeupper))
        (setq oldcont (getmem olditemptr))
        (setq cdrval (getmem (aplus2 olditemptr 1)))
        (while (wgreaterp olditemptr rangeend)
          (prog (tagoldcont)
           (setq clearedvalue (wand oldcont mask2))
           (setq tagoldcont (tag oldcont))
           (when (wlessp oldcont 0)
             (when (eq tagoldcont hskip)
                   (setq olditemptr (wdifference olditemptr
                                                (inf oldcont)))
                   (go whilenext))
             (when (eq tagoldcont negint-tag)
                   (setq clearedvalue (wor oldcont mask1)))  )
 
           (putmem newitemptr clearedvalue)
           (putmem (aplus2 newitemptr 1) cdrval)
           (setq olditemptr (wplus2 olditemptr -2))
           (setq newitemptr (wplus2 newitemptr -2))
      whilenext
           (setq oldcont (getmem olditemptr))
           (setq cdrval (getmem (aplus2 olditemptr 1)))
        nil))
))
 
(de gcerror (message p)
  (errorprintf "***** Fatal error during garbage collection")
  (errorprintf message p)
  (while t (abort)))      % final exit
 
(de known-free-space ()
  (wdifference (wplus2 heaptrapbound 10) heaplast))
 
(de gcmessage ()
 (prog (a1 a2 a3 a4 gleit)
   (setq gleit (longquotient (wtimes2 100 gctime*) (timc))) 
   (setq  a1  (car &oldstable&)
          a2  (cdr &oldstable&)
          a3  (wdifference heaplast localhole )
          a4  (wdifference localholeupper (wplus2 heaptrapbound 10)))
   (when (eq localhole (wdifference heaplowerbound 1))   %nothing freed
                (setq a3 0))
   (when (or (eq 0 localheapshrinkupper)
           (eq localholeupper (wplus2 heapupperbound 1)))%nothing freed
                (setq a4 0))
 
  (errorprintf "*** GC %w: time %d ms  at %w (~ %w ms job cpu time  gc : %w %%)"
                gcknt* startingrealtime (timeofday) (timc) gleit)
  (when (wlessp heaplast localhole)
       (console-print-string "Alles im Eimer")
       (setq *dump t) (center-dump heaplast 60)
                      (center-dump localhole 60))
  (when (eq *gc 'timing)
    (errorprintf "*** Timing %d %d %d %d %d %d %d %d %d"
       gctiming0 gctiming1 gctiming2 gctiming3 gctiming4 gctiming5
       gctiming6 gctiming7 gctiming8))
  (errorprintf
      "*** %d + %d recovered, %d + %d stable, %d + %d active, %d free"
           localheapshrinklower localheapshrinkupper a1 a2 a3 a4
           (intinf (known-free-space)))))
 
 
% the function cons-reclaim is called automatically by the cmacro
% *cons in case of heap overflow. The purpose of this routine
% is the saving of the current environment with all relevant
% registers, calling of the garbage collector and restoring of
% the environment.
 
(fluid '(*vector-register-save-array*))

(setq *vector-register-save-array* (gtwarray 512))

(compiletime (off counting))
 
(if_system Cray-2         % S0 contains callers A0 , must be saved
(lap '((*entry cons-reclaim expr 0)
       (*move (reg s0) (reg a1))             % A1 is destroyed anyway
       (*move ($Fluid tregsave) (reg a6))    % aux save area
       (store (reg s6) 0 (reg a6))           % s6 and s7 are destroyed
       (store (reg s7) 1 (reg a6))           % during alloc
       (move (reg s0) (reg a1))
       (store (reg S0) 2 (reg a6))           % S0 too on Cray-2
       (*alloc 62)
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))            % restore temp save S6 S7
       (load (reg s7) 1 (reg a6))
       (load (reg s0) 2 (reg a6))            % and s0
 
       (*call cons-reclaim-aux)
 
       (*move (reg s0) (reg a1))
       (*move ($fluid tregsave) (reg a6))
       (store (reg s6) 0 (reg a6))
       (store (reg s7) 1 (reg a6))
       (move (reg s0) (reg a1))
       (store (reg S0) 2 (reg a6))
       (*dealloc 62)
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))
       (load (reg s7) 1 (reg a6))
       (load (reg s0) 2 (reg a6))
       (*end 0)
)     )
(lap '((*entry cons-reclaim expr 0)
       (*move ($Fluid tregsave) (reg a6))    % aux save area
       (store (reg s6) 0 (reg a6))           % s6 and s7 are destroyed
       (store (reg s7) 1 (reg a6))           % during alloc
       (*move 64 (reg a6))             % saving all vregister 
       (*vsetvl (reg a6))
       (*move (reg a6) (reg s6))
       (*move ($fluid *vector-register-save-array*) (reg a6))
       (*move (reg a6) (reg s7))
       (*move (reg a6) (reg a0))
       (vstore (reg v0) (reg a0))
       (*wplus2 (reg s7) (reg s6))
       (*move (reg s7) (reg a0))
       (vstore (reg v1) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0))
       (vstore (reg v2) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0))
       (vstore (reg v3) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0))
       (vstore (reg v4) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0))
       (vstore (reg v5) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0))
       (vstore (reg v6) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0))
       (vstore (reg v7) (reg a0))      % all stored!
       (*alloc 33)
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))            % restore temp save S6 S7
       (load (reg s7) 1 (reg a6))
 
       (*call cons-reclaim-aux)
 
       (*move ($fluid tregsave) (reg a6))
       (store (reg s6) 0 (reg a6))
       (store (reg s7) 1 (reg a6))
       (*move 64 (reg a6))             % restoring all vregister  
       (*vsetvl (reg a6)) 
       (*move (reg a6) (reg s6)) 
       (*move ($fluid *vector-register-save-array*) (reg a6)) 
       (*move (reg a6) (reg s7)) 
       (*move (reg a6) (reg a0)) 
       (vload  (reg v0) (reg a0)) 
       (*wplus2 (reg s7) (reg s6)) 
       (*move (reg s7) (reg a0)) 
       (vload  (reg v1) (reg a0))  
       (*wplus2 (reg s7) (reg s6))  
       (*move (reg s7) (reg a0))
       (vload  (reg v2) (reg a0))
       (*wplus2 (reg s7) (reg s6))  
       (*move (reg s7) (reg a0)) 
       (vload  (reg v3) (reg a0))  
       (*wplus2 (reg s7) (reg s6))  
       (*move (reg s7) (reg a0)) 
       (vload  (reg v4) (reg a0))  
       (*wplus2 (reg s7) (reg s6))  
       (*move (reg s7) (reg a0)) 
       (vload  (reg v5) (reg a0))  
       (*wplus2 (reg s7) (reg s6))  
       (*move (reg s7) (reg a0)) 
       (vload  (reg v6) (reg a0))  
       (*wplus2 (reg s7) (reg s6))  
       (*move (reg s7) (reg a0)) 
       (vload  (reg v7) (reg a0))      % all restored! 
       (*dealloc 33)
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))
       (load (reg s7) 1 (reg a6))
       (*end 0)
)     )) 
 
(if_system Cray-2
(lap '((*entry cons-reclaim-aux  expr 0)
       (*move (reg s0) (reg a1))             % A1 is destroyed anyway
       (*move ($Fluid tregsave) (reg a6))    % aux save area
       (store (reg s6) 0 (reg a6))           % s6 and s7 are destroyed
       (store (reg s7) 1 (reg a6))           % during alloc
       (move (reg s0) (reg a1))
       (store (reg S0) 2 (reg a6))           % during alloc
       (*alloc 34)
       (*move (reg T1) (reg s6))             % T1 holds old S7 copy
       (*move (reg s6) (frame 15))
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))            % restore temp save S6 S7
       (load (reg s7) 1 (reg a6))
       (load (reg s0) 2 (reg a6))
       (*move (reg S0) (frame 29))           % save S regs
       (*move (reg  1) (frame 2))
       (*move (reg  2) (frame 3))
       (*move (reg  3) (frame 4))
       (*move (reg  4) (frame 5))
       (*move (reg  5) (frame 6))
       (*move (reg s6) (frame 7))
       (*move (reg s7) (frame 8))
       (*move (reg a7) (frame 1))
       (*move (reg a1) (frame 26))
       (*move (reg a2) (frame 27))
       (*move (reg a5) (frame 28))
       (*move (reg t126) (frame 9))          % save local T regs
       (*move (reg t127) (frame 10))
       (*move (reg t130) (frame 11))
       (*move (reg t131) (frame 12))
       (*move (reg t132) (frame 13))
       (*move (reg t133) (frame 14))
       (*move (reg t134) (frame 30))
       (*move (reg t135) (frame 31))
       (*move (reg t136) (frame 32))
       (*move (reg t137) (frame 33))
       (*move (reg t140) (frame 34))
       (*move (reg 6)   (frame 16))
       (*move (reg 7)   (frame 17))
       (*move (reg 8)   (frame 18))
       (*move (reg 9)   (frame 19))
       (*move (reg 10)  (frame 20))
       (*move (reg 11)  (frame 21))
       (*move (reg 12)  (frame 22))
       (*move (reg 13)  (frame 23))
       (*move (reg 14)  (frame 24))
       (*move (reg 15)  (frame 25))
 
       (*link !%sweep-stack expr 0)           % clean up a little bit
       (*link !%reclaim expr 0)               % call GC
 
       (*move (frame 16) (reg 6))
       (*move (frame 17) (reg 7))
       (*move (frame 18) (reg 8))
       (*move (frame 19) (reg 9))
       (*move (frame 20) (reg 10))
       (*move (frame 21) (reg 11))
       (*move (frame 22) (reg 12))
       (*move (frame 23) (reg 13))
       (*move (frame 24) (reg 14))
       (*move (frame 25) (reg 15))
       (*move (frame 1)  (reg a7))        % restore the whole stuff
       (*move (frame 26) (reg a1))
       (*move (frame 27) (reg a2))
       (*move (frame 28) (reg a5))
       (*move (frame 9)  (reg t126))
       (*move (frame 10) (reg t127))
       (*move (frame 11) (reg t130))
       (*move (frame 12) (reg t131))
       (*move (frame 13) (reg t132))
       (*move (frame 14) (reg t133))
       (*move (frame 30) (reg t134))
       (*move (frame 31) (reg t135))
       (*move (frame 32) (reg t136))
       (*move (frame 33) (reg t137))
       (*move (frame 34) (reg t140))
       (*move (frame 29) (reg S0))
       (*move (frame 2) (reg  1))
       (*move (frame 3) (reg  2))
       (*move (frame 4) (reg  3))
       (*move (frame 5) (reg  4))
       (*move (frame 6) (reg  5))
       (*move (frame 15) (reg s6))
       (*move (reg s6) (reg T1))
%
% restore reg s6 and s7 as last values
%
       (*move (frame 7) (reg s6))
       (*move (frame 8) (reg s7))
       (*move (reg s0) (reg a1))             % A1 is destroyed anyway
       (*move ($fluid tregsave) (reg a6))
       (store (reg s6) 0 (reg a6))
       (store (reg s7) 1 (reg a6))
       (move (reg s0) (reg a1))
       (store (reg S0) 2 (reg a6))
       (*dealloc 34)
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))
       (load (reg s7) 1 (reg a6))
       (load (reg s0) 2 (reg a6))
       (*end 0)
)     )
(lap '((*entry cons-reclaim-aux  expr 0)
       (*move ($Fluid tregsave) (reg a6))    % aux save area
       (store (reg s6) 0 (reg a6))           % s6 and s7 are destroyed
       (store (reg s7) 1 (reg a6))           % during alloc
       (*alloc 28)
       (*move (reg T1) (reg s6))             % T1 holds old S7 copy
       (*move (reg s6) (frame 15))
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))            % restore temp save S6 S7
       (load (reg s7) 1 (reg a6))
       (*move (reg  1) (frame 2))            % save S regs
       (*move (reg  2) (frame 3))
       (*move (reg  3) (frame 4))
       (*move (reg  4) (frame 5))
       (*move (reg  5) (frame 6))
       (*move (reg s6) (frame 7))
       (*move (reg s7) (frame 8))
       (*move (reg a7) (frame 1))           % a7 holds stacked B00
       (*move (reg a1) (frame 26))
       (*move (reg a2) (frame 27))
       (*move (reg a5) (frame 28))
       (*move (reg t71) (frame 9))          % save local T regs
       (*move (reg t72) (frame 10))
       (*move (reg t73) (frame 11))
       (*move (reg t74) (frame 12))
       (*move (reg t75) (frame 13))
       (*move (reg t76) (frame 14))
       (*move (reg 6)   (frame 16))
       (*move (reg 7)   (frame 17))
       (*move (reg 8)   (frame 18))
       (*move (reg 9)   (frame 19))
       (*move (reg 10)  (frame 20))
       (*move (reg 11)  (frame 21))
       (*move (reg 12)  (frame 22))
       (*move (reg 13)  (frame 23))
       (*move (reg 14)  (frame 24))
       (*move (reg 15)  (frame 25))
 
       (*link !%sweep-stack expr 0)           % clean up a little bit
       (*link !%reclaim expr 0)               % call GC
 
       (*move (frame 16) (reg 6))
       (*move (frame 17) (reg 7))
       (*move (frame 18) (reg 8))
       (*move (frame 19) (reg 9))
       (*move (frame 20) (reg 10))
       (*move (frame 21) (reg 11))
       (*move (frame 22) (reg 12))
       (*move (frame 23) (reg 13))
       (*move (frame 24) (reg 14))
       (*move (frame 25) (reg 15))
       (*move (frame 1)  (reg a7))          % restore the whole stuff
       (*move (frame 26) (reg a1))
       (*move (frame 27) (reg a2))
       (*move (frame 28) (reg a5))
       (*move (frame 9)  (reg t71))
       (*move (frame 10) (reg t72))
       (*move (frame 11) (reg t73))
       (*move (frame 12) (reg t74))
       (*move (frame 13) (reg t75))
       (*move (frame 14) (reg t76))
       (*move (frame 2) (reg  1))
       (*move (frame 3) (reg  2))
       (*move (frame 4) (reg  3))
       (*move (frame 5) (reg  4))
       (*move (frame 6) (reg  5))
       (*move (frame 15) (reg s6))
       (*move (reg s6) (reg T1))
%
% restore reg s6 and s7 as last values
%
       (*move (frame 7) (reg s6))
       (*move (frame 8) (reg s7))
       (*move ($fluid tregsave) (reg a6))
       (store (reg s6) 0 (reg a6))
       (store (reg s7) 1 (reg a6))
       (*dealloc 28)
       (*move ($fluid tregsave) (reg a6))
       (load (reg s6) 0 (reg a6))
       (load (reg s7) 1 (reg a6))
       (*end 0)
)     )) 
 
(compiletime (on  counting))
 
(de !%sweep-stack()
   (prog (stackp stack-low tg x info hpa)
     (setq hpa heap)
     (setq stackp st)
     (if_system Cray-2 (setq stack-low (wplus2 stackp -99))
                       (setq stack-low (wplus2 stackp -69)))
     (ifor (from i stackp stack-low -1)
       (do (prog ()
        (setq x (getmem i))
        (setq tg (tag x))
        (setq info (inf x))
        (case tg
           ((id-tag) (when (or (wlessp info 0) (wgeq info maxsymbols))
                            (go killitem)))
           ((halfwords-tag) (go killitem)) % NO halfwords used
           ((fixnum-tag bignum-tag floatnum-tag words-tag)
                  (cond ((wgreaterp info heaplast) (go killitem))
                        ((wlessp info hpa) (go next)) %no high tags
                        ((not (eq (tag (getmem info)) hwords-tag))
                               (go killitem))))
           ((string-tag bytes-tag)
                   (cond ((wgreaterp info heaplast) (go killitem))
                         ((wlessp info hpa) (go next)) %no high tags
                         ((not (eq (tag (getmem info)) hbytes-tag))
                               (go killitem))))
           ((vector-tag evector-tag)
                   (cond ((wgreaterp info heaplast) (go killitem))
                         ((wlessp info hpa) (go next)) %no high tags
                         ((not (eq (tag (getmem info)) hvect-tag))
                              (go killitem))))
           ((pair-tag)
                   (cond ((wgreaterp info heapupperbound) (go killitem)))
                   (setq tg2 (tag (getmem info)))
                   (cond ((eq tg2 negint-tag) t)
                         ((wlessp tg2 16) t)
                         (t (go killitem)))
                   (setq tg2 (tag (getmem (aplus2 info 1))))
                   (cond ((eq tg2 negint-tag) t)
                         ((wlessp tg2 16) t)
                         (t (go killitem))))
            (nil nil))
     (go next)
   killitem (putmem i 0)
            (when *dump
             (console-print-string "+++ Following item deleted")
             (putoct x) (terminalwritechar 32 32)
             (putoct i) (terminalwritechar 32 32)
             (putoct stackp) (terminalwritechar 32 32)
             (console-newline))
   next
 ) ) ) ))
