%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:SYSTEM-IO.SL                                            
% Description:  HP-UX specific IO routines for PSL                         
% Author:       Russ Fish, Robert R. Kessler, Martin L. Griss              
% Created:      4-May-84                                                   
% Modified:     05-Oct-87 (Harold Carr)
% Mode:         Lisp                                                       
% Package:                                                                 
% Status:       Experimental (Do Not Distribute)                           
%
% (c) Copyright 1987, University of Utah, all rights reserved.        
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 09-Jan-91 (Herbert Melenk)
%  definition for unixnull and clearerr for MS-DOS
% 03-May-90 (Herbert Melenk)
%  inserted file name expansion for MS-DOS type environments
% 05-Oct-87 (Harold Carr)
%  Added call to GETUNIXARGS to SYSCLEARIO to make consistent with VAX.
% 31-Mar-87 (Leigh Stoller)
%  Added calls to setlinebuf back in. This should be in all systems just in
%   line buffering is a problem.
% 15-Feb-85 13:27:04 (Brian Beach)
%  Commented out the setlinebuf calls in syscleario.
% 17-Dec-84 11:14:32 (Vicki O'Day)
%  Uncommented out the setlinebuf calls in syscleario.
% 10-Jul-84 13:30:58 (RAM)
%  Replaced calls to fopen with call to unixopen.  Unixopen will expand
%  shell variables (and some other fancy stuff).
% 21-May-84 (Vicki O'Day)                                                  
%  Added unixcleario call to syscleario.  This is so buffers that are      
%  dirty when a savesystem is done will be cleared.                        
% 4-May-84 11:16:33 (Brian Beach)                                          
%  Added VAXREADCHAR and VAXWRITECHAR to the version received from Utah.   
%  These are needed for RAWIO.                                             
%  Translated from RLISP to LISP.                                          
% Edit by Russ Fish, Wed Dec 21 17:10:07 1983                              
%  Force line buffering on StdOut and StdErr in SysClearIO even if file output.
%                                                                          
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Each individual system must have the following routines defined.         
% SysClearIo, SysOpenRead, SysOpenWrite, SysReadRec, SysWriteRec, SysClose,
% SysMaxBuffer                                                             
%                                                                          
%   The following definitions are used in the routines:                    
%    FileDescriptor - A machine dependent word that references a file once 
%                     opened.                                              
%    FileName - A Lisp string of the file name.                            
% FileDescriptors in the Unix version are "FILE *" (pointers to C "standard
% I/O" stream structs, as defined in /usr/include/stdio.h.)                

(compiletime (load io-decls sys-macros fast-vector))

(on fast-integers)

(fluid '(unix-stderr* unix-tty* **windows))

(de clearerr(dummy) nil)

%   C support code in unix-io.c imports knowledge of stdio.h .             
(fluid '(unixeof unixnull unixstderr unixstdout unixstdin unixtty))

(setq unixnull 0)

% Managed here.                                                            
(de unixstring (s)
  % Strip tags for passing a string.                                       
  (if (stringp s)
    (strbase (strinf s))
    (stderror (bldmsg "Not a string: %p%n" s))))

% Use the C/stdio strlen fn
(de unixstrlen (s)
  % Count chars until null (0) char.                                      
  (external_strlen (unixstring s)))

%------------------------------------------------------------------------------
									   
% Any system dependent clearing that has to be done.                       
(de syscleario ()
  % Establish channels #'s for initial system.                             
  (setf stdin* 0)
  (setf stdout* 1)
  (setf errout* stdout*)
  (setf promptout* stdout*)
  % Hand Build channeltables.                                              
  (if (not (weq **windows 1))
    (progn  
      (clearonechannel stdin* stdin*
		      'input 'terminalinputhandler)
     % With prompts.                                                          
     (clearonechannel stdout* stdout*
		      'output 'independentwritechar)
     (setf unix-stderr* 5)
     % As a system-dependent feature /csp                                     
     (clearonechannel unix-stderr*  unix-stderr* 
		      'output 'independentwritechar)
     (setf unix-tty* 6)
     % Controlling tty, even when stdin redirected.                           
     (clearonechannel unix-tty* unix-tty* 
		      'input 'terminalinputhandler)
    )                                                      
    (progn  
	 % section for WINDOWS
      (clearonechannel stdin* stdin*
		      'input 'terminalinputhandler)
     (clearonechannel stdout* stdout*
		      'output 'WINDOWSwritechar)
     (setf unix-stderr* 5)
     % As a system-dependent feature /csp                                     
     (clearonechannel unix-stderr*  unix-stderr* 
		      'output 'WINDOWSwritechar)
     (setf unix-tty* 6)
     % Controlling tty, even when stdin redirected.                           
     (clearonechannel unix-tty* unix-tty* 
		      'input 'terminalinputhandler)
   ))                                                      
   

  % Force line buffering on StdOut and StdErr even if file or pipe output. 
  % This only works here on BSD systems. SYSV must have the call in the
  % ukernel.

  %%(setlinebuf unixstdout)
  %%(setlinebuf unixstderr) 

  %%%(getunixargs)      
  nil)

%------------------------------------------------------------------------------
									   
% Open FileName for input and return a file descriptor used in later       
% references to the file.                                                  
(de sysopenread (channel filename)
  (prog (filed)
	% Open a Unix StdIO channel.                                       
	(setq filename (fnexpand filename))
	(&time-control nil)
	(setq filed (unixopen (unixstring filename) (unixstring "r")))
	(&time-control t)
	% unixOpen returns NULL if there is an error.                         
	(return (if (weq filed 2)
		  (contopenerror filename 'input)
		  filed))))

%------------------------------------------------------------------------------
									   
% Open FileName for output and return a file descriptor used in later      
% references to the file.                                                  
(de sysopenwrite (channel filename)
  (prog (filed)
	% Open a Unix StdIO channel.                                       
	(setq filename (fnexpand filename)) 
	(&time-control nil)
	(setq filed (unixopen (unixstring filename) (unixstring "w")))
	(&time-control t)
	% unixOpen returns NULL if there is an error.                         
	(return (if (weq filed 2)
		  (contopenerror filename 'output)
		  filed))))

%------------------------------------------------------------------------------
									   
% Read from the FileDescriptor, putting a line record into the StringBuffer
% terminated by an EOL or EOF character.  Return the length of the string read
% as the indexing bound (one less than the number of characters read.)     

(de sysreadrec (filedescriptor stringbuffer)
    (prog (l r)

	  (&time-control nil)
		   % Ensure prompts seen before input. 
	  (when (or (weq filedescriptor unixstdin) 
		    (weq filedescriptor unixtty))
		(fflush unixstdout))
	  (setq r (fgets (unixstring stringbuffer)     
			 (sysmaxbuffer filedescriptor) filedescriptor))                 
	  (&time-control t)

	  (cond ((wneq r unixnull) 
		 % Check that not end of file.                             
		 (setq l (wdifference (unixstrlen stringbuffer) 1))
		   % DOS: cancel addtional eol's on standard input
		 (when (weq filedescriptor 0)
		    (while (and (wgreaterp l 0)
				(weq (strbyt (strinf stringbuffer) l) (char eol)))
			   (setq l (wdifference l 1))))
		 (return l))
		(t                                        
		   (progn % Return an EOF character on end of file.    
			  (setf (strbyt (strinf stringbuffer) 0) $eof$)
			  (clearerr filedescriptor)
			  % TTYs can type again after EOF.                   
			  (return 0))))))

    % Index of EOF char.
%----------------------------------------------------------------------------
									   
% Write StringLength+1 characters from StringToWrite from the first position.  
									   
(de syswriterec (filedescriptor stringtowrite stringlength)
  (&time-control nil)
  (fwrite (unixstring stringtowrite) (wplus2 stringlength 1) 1
	  filedescriptor)
  (&time-control t)          
)

%-----------------------------------------------------------------------------
									   
%   Close FileDescriptor, allowing it to be reused.                        
(de sysclose (filedescriptor)
  (&time-control nil)
  (fclose filedescriptor)
  (&time-control t)
  )

%-----------------------------------------------------------------------------
									   
%   Return the maximum buffer size.                                        
(de sysmaxbuffer (filedesc)
  200)

%------------------------------------------------------------------------------
									   
(de contopenerror (filename accessmode)
  (continuableerror 99 
		    (bldmsg "`%s' cannot be open for %w" filename 
		     accessmode)
		    (list 'open (mkstr filename) (mkquote accessmode))))

%
% using a buffered variant instead (variant of independentwritechar)
 
(de WINDOWSwritechar (channel chr)
 
 (when (not (equal chr (char cr)))  % inhibit isolated cr's
  (or (eq channel 1) (testlegalchannel channel))
  (when (equal chr (char eol))
      (setf (wgetv nextposition channel) (+ (wgetv nextposition channel) 1))
      (setf (strbyt (strinf (igetv iobuffer channel))
		(wgetv nextposition channel)) 13) % CR
      )
  (setf (wgetv nextposition channel) (+ (wgetv nextposition channel) 1))
  (setf (strbyt (strinf (igetv iobuffer channel))
		(wgetv nextposition channel))
    chr)
  (when (or (equal chr (char eol))
	(>= (iadd1 (wgetv nextposition channel))   % Indexed from 0
	    (wgetv bufferlength channel)))        % Number of characters
    (WINDOWSflushbuffer channel)
  ) )
  )

(de WINDOWSflushbuffer(channel)
  (setf (strbyt (strinf (igetv iobuffer channel))
		(iadd1 (wgetv nextposition channel)))
    0)
  (&time-control nil)
  (psll-call (strbase (strinf (igetv iobuffer channel))) 0 0 1)  % puts call 
  (&time-control t)
  (setf (wgetv nextposition channel) -1)
  )

   
% End of file.
